/******************************************************
*         PARALLEL RANDOM NUMBER GENERATOR            *
*       USING SHIFT-REGISTER SEQUENCE METHOD          *
*                                                     *
*                      by                             *
*                                                     *
*                Ting-Wai Chiu        Sept,1987.      *
*                                                     *
*    Changed by Paul Coddington so that it works,     *
*    Oct, 1990.                                       *
******************************************************/

#include "comext.h"
#include <math.h>
#include <stdio.h>

#define A 3141592621 
#define C 137

#define PT0_INIT    0x00dd    /* Initialize pointers for TWC rng    */
#define PT147_INIT  0x004a    /* (instead of reading from pts file) */

long  int seeds[250];
int   pt0,pt147;
double twoto31 = 2147483648.0;
char seedname[128];

void initseed()           /* generate the initial seeds and pointers */
{
  FILE *f1,*f2,*fopen();
  register long int tmp;
  long int tmp1;
  register short int i,j;

  /* Initialize seeds using Prand */

  tmp = iranf();
  for (i=249;i>=0;i--)
  {
    tmp1 = iranf();
    tmp = (tmp * A + tmp1) & 0x7fffffff;
    seeds[i] = tmp^tmp1;
  }

  /* every node gets the same pointers */
  pt0 = PT0_INIT;
  pt147 = PT147_INIT;
}

void seed(rw)
int rw ;
{
  FILE *f1,*f2,*fopen();
  long extick();
  register long int tmp ;
  register short int i ;
  int sz;
  sz = 4096;

  if (rw == 1)              /* read the initial seeds and pointers */
  {
    sprintf(seedname, "seeds.%s", lastfname);
    f1 = fopen(seedname,"r");
    fmulti(f1);
    fread(seeds,sizeof(long),250,f1);
    fclose(f1);
    sprintf(seedname, "pts.%s", lastfname);
    f2 = fopen(seedname,"r");
    fmulti(f2);
    fscanf(f2,"%04x %04x",&pt0,&pt147);
    fclose(f2);
  }

  else if ( rw == 0 )        /* save the current seeds and pointers */
  {
    sprintf(seedname, "seeds.%s", fname);
    f1 = fopen(seedname,"w") ;
    fmulti(f1);
    fwrite(seeds,sizeof(long),250,f1);
    fclose(f1);
    sprintf(seedname, "pts.%s", fname);
    f2 = fopen(seedname,"w");
    fmulti(f2);
    fprintf(f2,"%04x %04x\n",pt0,pt147);
    fflush(f2);
    fclose(f2);
  }

  else if ( rw == 2 )        /* randomize the seeds and pointers */
  {
    f1 = fopen("seeds","r");
    fmulti(f1);
    fread(seeds,sizeof(long),250,f1);
    for (i=249;i>=0;i--)
    {
      tmp = iranf();
      tmp = (tmp * 16807 ) & 0x7fffffff ;
      seeds[i] ^= tmp ;
    }
    fclose(f1);
    f1 = fopen("seeds","w") ;
    fmulti(f1);
    fwrite(seeds,sizeof(long),250,f1);
    fclose(f1);
    f2 = fopen("pts","r");
    fmulti(f2);
    fscanf(f2,"%04x %04x",&pt0,&pt147);
    fclose(f2);
    f2 = fopen("pts","w");
    fmulti(f2);
    tmp = iranf();
    tmp = (tmp * 16807 ) & 0x7fffffff ;
    pt0 = (pt0 + (int) tmp) % 250 ;
    pt147 = (pt147 + (int) tmp ) % 250 ;
    fprintf(f2,"%04x %04x\n",pt0,pt147);
    fflush(f2);
    fclose(f2);
  }
  else
    printf("\n The seed(rw) must be called with rw = 0, 1, or 2 \n");
}

float pranf()
{
  register long int tmp ;
  register int t0,t147 ;
  t0 = pt0 ;
  t147 = pt147 ;
  tmp = seeds[t147] ;
  /* Addition is better than OR */
  /* seeds[t0] ^= tmp ; */
  seeds[t0] = (seeds[t0] + tmp) & 0x7fffffff ;
  if (--t0 < 0)
    t0 = 249;
  if (--t147 < 0)
    t147 = 249;
  pt0 = t0 ;
  pt147 = t147 ;
  return( (float) ( (double) tmp / twoto31)) ;
}

float normal()
{
  register double x ;
  register long int tmp ;
  register int t0,t147 ;
  register char loop ;
  t0 = pt0 ;
  t147 = pt147 ;
  x = 0 ;
  for (loop=12;loop!=0;loop--)
  {
    tmp = seeds[t147] ;
    /* Addition is better than OR */
    /* seeds[t0] ^= tmp ; */
    seeds[t0] = (seeds[t0] + tmp) & 0x7fffffff ;
    x += tmp ;
    if (--t0 < 0)
      t0 = 249 ;
    if (--t147 < 0)
      t147 = 249 ;
  }
  pt0 = t0 ;
  pt147 = t147 ;
  return( (float) (ldexp( (double) x, (int) -31) - 6)) ;
}

int ipranf(imax)
int imax ;
{
  register long int tmp ;
  register int t0,t147 ;
  t0 = pt0 ;
  t147 = pt147 ;
  tmp = seeds[t147] ;
  /* Addition is better than OR */
  /* seeds[t0] ^= tmp ; */
  seeds[t0] = (seeds[t0] + tmp) & 0x7fffffff ;
  if (--t0 < 0)
    t0 = 249 ;
  if (--t147 < 0)
    t147 = 249 ;
  pt0 = t0 ;
  pt147 = t147 ;
  return(tmp%imax) ;
}

